#include <stdio.h>
#include <string>
#include <sstream>
#include "Rectangle.h"
#include "json.h"
#include "Writer.h"
#include "Processor.h"
#include "Visualizer.h"
#include "NormalProcessor.h"
#include "SourceProcessor.h"
#include "BoundaryProcessor.h"
#include "RulesTableProcessor.h"

extern long field_width, field_height, iters, end_state;
extern std::vector<Writer*> Writers;
extern std::vector<Processor*> Procs;

std::string loadFileToString(const char* filename)
{
	std::string result;

	FILE* f = fopen(filename, "r");
	if(f==NULL)
	{
		std::cerr << "Failed to open file: " << filename << " for reading." << std::endl;
		abort();
	}
	while(!feof(f))
	{
		char c;
		if(fscanf(f, "%c", &c)==1)
			result += c;
	}
	fclose(f);

	return result;
}

void load_global_settings(const json_value& root)
{
	field_width = (long)((root)["field_width"]);
	field_height = (long)((root)["field_height"]);
	iters = (long)((root)["iters"]);

	{
		long rp_bits = (long)((root)["rest_particles_bits"]);
		end_state = 16;
		while(rp_bits-->0)
			end_state *= 2;
	}
}

static Rectangle get_rectangle(const json_value& val)
{
	long left, top, right, bottom;
	left = val["area"][0];
	top = val["area"][1];
	right = val["area"][2];
	bottom = val["area"][3];
	assert(left < right && top < bottom);
	assert(left>=0 && top>=0);

	return Rectangle(left, top, right-left, bottom-top);
}

static Processor* load_normal_processor(const json_value& mod)
{
	Rectangle area = get_rectangle(mod);
	bool check_probs = mod["check_probabilities"];
	const char* save_table = mod["save_table"];
	bool no_rp_create = mod["no_rp_create"];
	NormalProcessor* p = new NormalProcessor(end_state, area, check_probs, save_table,
			no_rp_create);
	return p;
}

static Processor* load_rulestable_processor(const json_value& mod)
{
	Rectangle area = get_rectangle(mod);
	std::string file = (const char*)mod["file"];
	long end_state = mod["end_state"];

	return new RulesTableProcessor(file, area, end_state);
}

Processor* load_constsource_processor(const json_value& mod)
{
	Rectangle area = get_rectangle(mod);
	const json_value &states = mod["states"];
	std::vector<std::pair<float, int> > src_states;
	for(int j=0; j<states.u.array.length; j++)
	{
		float prob = (double)states[j][0];
		long st = states[j][1];
		src_states.push_back(std::make_pair(prob, (int)st));
	}
	const json_value &act_int = mod["active_intervals"];
	std::vector<std::pair<int, int> > intervals;
	for(int j=0; j<act_int.u.array.length; j++)
	{
		long start, end;
		start = act_int[j][0];
		end = act_int[j][1];
		intervals.push_back(std::make_pair(start, end));
	}
	return new SourceProcessor(area, src_states, intervals);
}

Processor* load_circonstsource_processor(const json_value& mod)
{
	long x = mod["coord"][0],
		 y = mod["coord"][1],
		 r = mod["radius"];
	const json_value &states = mod["states"];
	std::vector<std::pair<float, int> > src_states;
	for(int j=0; j<states.u.array.length; j++)
	{
		float prob = (double)states[j][0];
		long st = states[j][1];
		src_states.push_back(std::make_pair(prob, (int)st));
	}
	const json_value &act_int = mod["active_intervals"];
	std::vector<std::pair<int, int> > intervals;
	for(int j=0; j<act_int.u.array.length; j++)
	{
		long start, end;
		start = act_int[j][0];
		end = act_int[j][1];
		intervals.push_back(std::make_pair(start, end));
	}
	return new CircleSourceProcessor(x, y, r, src_states, intervals);
}

Processor* load_randomsource_processor(const json_value& mod)
{
	Rectangle area = get_rectangle(mod);
	long end_state = mod["end_state"];
	const json_value &act_int = mod["active_intervals"];
	std::vector<std::pair<int, int> > intervals;
	for(int j=0; j<act_int.u.array.length; j++)
	{
		long start, end;
		start = act_int[j][0];
		end = act_int[j][1];
		intervals.push_back(std::make_pair(start, end));
	}
	return new RandomSourceProcessor(area, end_state, intervals);
}

void load_modifiers(const json_value& mods)
{
	for(int i=0; i<mods.u.array.length; i++)
	{
		const json_value &mod = mods[i];
		
		if(strcmp(mod["type"], "normal")==0)
			Procs.push_back(load_normal_processor(mod));
		else if(strcmp(mod["type"], "barrier")==0)
			Procs.push_back(new BoundaryProcessor(get_rectangle(mod)));
		else if(strcmp(mod["type"], "const-source")==0)
			Procs.push_back(load_constsource_processor(mod));
		else if(strcmp(mod["type"], "circle-const-source")==0)
			Procs.push_back(load_circonstsource_processor(mod));
		else if(strcmp(mod["type"], "random-source")==0)
			Procs.push_back(load_randomsource_processor(mod));
		else if(strcmp(mod["type"], "rules-table")==0)
			Procs.push_back(load_rulestable_processor(mod));
		else
		{
			std::cerr << "Unknown modifier type: " << (const char*)mod["type"] << std::endl;
			abort();
		}
	}
}

Writer* load_image_output(const json_value& out)
{
	long period = out["period"];
	long average_radius = out["average_radius"];
	std::string folder = (const char*)out["folder"];
	{
		std::ostringstream os; 
		os << "mkdir " << folder << " > /dev/null 2> /dev/null";
		int res=system(os.str().c_str());
	}
	std::vector<std::string> data;
	for(int j=0; j<out["data"].u.array.length; j++)
		data.push_back((const char*)out["data"][j]);
	return new Visualizer(period, average_radius, Visualizer::Image, folder, data,
			get_rectangle(out));
}

Writer* load_text_output(const json_value& out)
{
	long period = out["period"];
	long average_radius = out["average_radius"];
	std::string folder = (const char*)out["folder"];
	{
		std::ostringstream os; 
		os << "mkdir " << folder << " > /dev/null 2> /dev/null";
		int res=system(os.str().c_str());
	}
	std::vector<std::string> data;
	data.push_back((const char*)out["data"]);
	return new Visualizer(period, average_radius, Visualizer::Text, folder, data,
			get_rectangle(out));
}

Writer* load_dump_output(const json_value& out)
{
	long period = out["period"];
	std::string folder = (const char*)out["folder"];
	{
		std::ostringstream os; 
		os << "mkdir " << folder << " > /dev/null 2> /dev/null";
		int res=system(os.str().c_str());
	}
	return new DumpWriter(period, folder);
}


void load_outputs(const json_value& outs)
{
	for(int i=0; i<outs.u.array.length; i++)
	{
		const json_value& out = outs[i];
		long period = out["period"];
		if(strcmp(out["type"], "image-bmp")==0)
		{
			Writers.push_back(load_image_output(out));
		}
		else if(strcmp(out["type"], "text-log")==0)
		{
			Writers.push_back(load_text_output(out));
		}
		else if(strcmp(out["type"], "echo-iter")==0)
		{
			Writers.push_back(new IterWriter(period, iters));
		}
		else if(strcmp(out["type"], "dump")==0)
		{
			Writers.push_back(load_dump_output(out));
		}
		else
		{
			std::cerr << "Unknown writer type: " << (const char*)out["type"] << std::endl;
			abort();
		}
	}
}

void load_configuration(const char* filename)
{
	std::string config = loadFileToString(filename);

	json_value* root = json_parse(config.c_str());

	if(root == NULL)
	{
		std::cerr << "Error: file " << filename << " is not a valid JSON file" << std::endl;
		abort();
	}
	
	assert(root != NULL);
	assert(root->type == json_object);

	try
	{
		load_global_settings(*root);
		load_modifiers((*root)["modifiers"]);
		load_outputs((*root)["outputs"]);
	}
	catch(json_value::NotFoundException)
	{
		std::cerr << "Error: expected field in configuration file " << filename << " was not found" << std::endl;
		abort();
	}
}

